/*
 * yasapi_engine_common.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2018  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>
#include <stdint.h>

///////////////////////////////////////////////////////////////////////////////
#if ! defined (AUDCLNT_STREAMFLAGS_AUTOCONVERTPCM) // [
#define AUDCLNT_STREAMFLAGS_AUTOCONVERTPCM 0x80000000
#endif // ]

#if ! defined (AUDCLNT_STREAMFLAGS_SRC_DEFAULT_QUALITY) // [
#define AUDCLNT_STREAMFLAGS_SRC_DEFAULT_QUALITY 0x08000000
#endif // ]

///////////////////////////////////////////////////////////////////////////////
static int EngineCreateBase(Engine *pEngine, Player *pPlayer);
static int EngineCreateOptions(Engine *pEngine);
static int EngineGetBufferDuration(Engine *pEngine);
static int EngineCreateRing(Engine *pEngine, DWORD dwChunkSizeSource,
    ERingAlloc eRingAlloc);

int EngineCreate(Engine *pEngine, Player *pPlayer, int samplerate,
    int numchannels, int bitspersamp)
{
  int nMaxLatency=0;
  const EEngine eState=pEngine->eState;

  int bSurround=pEngine->options.common.bSurround;
  WORD nChannelsTarget;
  WORD wBitsPerSampleTarget;
  Wasapi *pWasapi;

  /////////////////////////////////////////////////////////////////////////////
  if (eState<eEngineBase&&EngineCreateBase(pEngine,pPlayer)<0) {
    DMESSAGE("creating engine base");
    EngineKill(pEngine,eState);
    goto exit;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (!(pWasapi=PlayerWasapi(pPlayer))) {
    DMESSAGE("accessing WASAPI");
    EngineDestroy(pEngine);
    goto exit;
  }

  pEngine->pDevice=pWasapi->selected.pDevice->pDevice;

  /////////////////////////////////////////////////////////////////////////////
  ZeroMemory(&pEngine->param,sizeof pEngine->param);

  /////////////////////////////////////////////////////////////////////////////
  if (pEngine->options.common.bLatency)
    pEngine->param.time.lpVtbl=&gcIEngineTimeNullVtbl;
  else
    pEngine->param.time.lpVtbl=&gcIEngineTimeClockVtbl;

  /////////////////////////////////////////////////////////////////////////////
  ZeroMemory(&pEngine->stats,sizeof pEngine->stats);
  pEngine->stats.uNumTracks=1u;
  ++pEngine->pPlayer->stats.uNumTracks;
  pEngine->param.time.lpVtbl->Set(pEngine,0);

  if (EngineCreateOptions(pEngine)<0) {
    DMESSAGE("creating options");
    EngineKill(pEngine,eState);
    goto exit;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (WfxCreate(&pEngine->wfx.source,samplerate,numchannels,bitspersamp,
      bSurround)<0) {
    DVMESSAGE("unknown format %d/%d/%d",samplerate,numchannels,
        bitspersamp);
    EngineKill(pEngine,eState);
    goto exit;
  }

  if (1==numchannels&&pEngine->options.common.eMono2Stereo)
    nChannelsTarget=2;
  else
    nChannelsTarget=numchannels;

  switch (pEngine->options.device.eBitdepth) {
  case eBitdepth16:
    wBitsPerSampleTarget=16u;
    break;
  case eBitdepth24:
    wBitsPerSampleTarget=24u;
    break;
  case eBitdepth32:
    wBitsPerSampleTarget=32u;
    break;
  default:
    wBitsPerSampleTarget=bitspersamp;
    break;
  }

  if (WfxCreate(&pEngine->wfx.target,samplerate,nChannelsTarget,
      wBitsPerSampleTarget,bSurround)<0) {
    DVMESSAGE("unknown format %d/%u/%u",samplerate,nChannelsTarget,
        wBitsPerSampleTarget);
    EngineKill(pEngine,eState);
    goto exit;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (EngineCreateAudioClient(pEngine,eEngineAudioClientInititial)<0) {
    DMESSAGE("creating audio client");
    EngineKill(pEngine,eState);
    goto exit;
  }

  ++pEngine->eState;
  DWRITELN(EngineState(pEngine));
  DASSERT_ENGINE_STATE(pEngine,eEngineAudioClientInititial,exit);

  /////////////////////////////////////////////////////////////////////////////
  GtkUpdateDialog(&pPlayer->gtk,&pPlayer->gtk.config);
  nMaxLatency=pEngine->options.device.xMaxLatencyAssumed+0.5;
exit:
  return nMaxLatency;
}

static int EngineCreateBase(Engine *pEngine, Player *pPlayer)
{
  const EEngine eState=pEngine->eState;

  /////////////////////////////////////////////////////////////////////////////
  pEngine->pPlayer=pPlayer;

  /////////////////////////////////////////////////////////////////////////////
  srand(time(NULL));

  /////////////////////////////////////////////////////////////////////////////
  InitializeCriticalSection(&pEngine->cs);
  ++pEngine->eState;
  DASSERT(eEngineCriticalSection==pEngine->eState,exit);

  /////////////////////////////////////////////////////////////////////////////
  if (!(pEngine->timer.hEvent=CreateEventW(NULL,FALSE,FALSE,NULL))) {
    DMESSAGE("creating timer event");
    EngineKill(pEngine,eState);
    goto exit;
  }

  ++pEngine->eState;
  DASSERT(eEngineTimerEvent==pEngine->eState,exit);

  /////////////////////////////////////////////////////////////////////////////
  if (!(pEngine->timer.hTimer=CreateWaitableTimerW(NULL,FALSE,NULL))) {
    DMESSAGE("creating waitable timer");
    EngineKill(pEngine,eState);
    goto exit;
  }

  ++pEngine->eState;
  DASSERT(eEngineTimer==pEngine->eState,exit);

  /////////////////////////////////////////////////////////////////////////////
  if (!(pEngine->thread.hProducer=CreateEventW(NULL,FALSE,FALSE,NULL))) {
    DMESSAGE("creating producer event");
    EngineKill(pEngine,eState);
    goto exit;
  }

  ++pEngine->eState;
  DASSERT(eEngineEventProducer==pEngine->eState,exit);

  /////////////////////////////////////////////////////////////////////////////
  if (!(pEngine->thread.hConsumer=CreateEventW(NULL,FALSE,FALSE,NULL))) {
    DMESSAGE("creating consumer event");
    EngineKill(pEngine,eState);
    goto exit;
  }

  ++pEngine->eState;
  DASSERT(eEngineEventConsumer==pEngine->eState,exit);



  /////////////////////////////////////////////////////////////////////////////
  EnterCriticalSection(&pEngine->cs); // {
  pEngine->thread.pCommand=NULL;

  pEngine->thread.hThread=CreateThread(
    NULL,     // _In_opt_  LPSECURITY_ATTRIBUTES  lpThreadAttributes,
    0ul,      // _In_      SIZE_T                 dwStackSize,
    EngineThreadProc,
              // _In_      LPTHREAD_START_ROUTINE lpStartAddress,
    pEngine,  // _In_opt_  LPVOID                 lpParameter,
    0ul,      // _In_      DWORD                  dwCreationFlags,
    NULL      // _Out_opt_ LPDWORD                lpThreadId
  );

  if (!pEngine->thread.hThread) {
    DMESSAGE("creating thread");
    EngineKill(pEngine,eState);
    goto exit;
  }

  ++pEngine->eState;
  DASSERT(eEngineThread==pEngine->eState,exit);

  while (pEngine->eState<eEngineThreadRunning) {
    LeaveCriticalSection(&pEngine->cs); // {
    WaitForSingleObject(pEngine->thread.hConsumer,INFINITE);
    EnterCriticalSection(&pEngine->cs); // }
  }

  pEngine->thread.pCommand=NULL;
  LeaveCriticalSection(&pEngine->cs); // }


  /////////////////////////////////////////////////////////////////////////////
exit:
  DWRITELN(EngineState(pEngine));
  return eState<pEngine->eState?0:-1;
}

static int EngineCreateOptions(Engine *pEngine)
{
  int code=-1;
  OptionsCommon *pCommon=&pEngine->options.common;
  OptionsDevice *pDevice=&pEngine->options.device;
  Wasapi *pWasapi;

  /////////////////////////////////////////////////////////////////////////////
  if (!(pWasapi=PlayerWasapi(pEngine->pPlayer))) {
    DMESSAGE("accessing WASAPI");
    EngineDestroy(pEngine);
    goto exit;
  }

  DVWRITELN("=== %d =================================",
      pWasapi->selected.pDevice-pWasapi->pDevices);
  *pCommon=pWasapi->options.common.current;
  *pDevice=pWasapi->selected.pDevice->pMember->current;
  DVWRITELN("=== %s",Bitdepth(pDevice->eBitdepth));

  /////////////////////////////////////////////////////////////////////////////
  pEngine->param.pFriendlyName
      =pWasapi->selected.pDevice->pFriendlyNameOriginal;

  if (pDevice->eStrategy) {
    pEngine->param.dwStreamFlags=AUDCLNT_STREAMFLAGS_EVENTCALLBACK;
    DWRITELN("AUDCLNT_STREAMFLAGS_EVENTCALLBACK");
  }
  else
    pEngine->param.dwStreamFlags=0ul;

  if (pDevice->eMode) {
    pEngine->param.eShareMode=AUDCLNT_SHAREMODE_EXCLUSIVE;
    DWRITELN("AUDCLNT_SHAREMODE_EXCLUSIVE");
  }
  else {
    pEngine->param.eShareMode=AUDCLNT_SHAREMODE_SHARED;
    DWRITELN("AUDCLNT_SHAREMODE_SHARED");

    if (!pDevice->bAutoConvertPCM)
      goto shared;

    pEngine->param.dwStreamFlags|=AUDCLNT_STREAMFLAGS_AUTOCONVERTPCM;
    DWRITELN("AUDCLNT_STREAMFLAGS_AUTOCONVERTPCM");

    if (! pDevice->bSRCDefaultQuality)
      goto shared;

    pEngine->param.dwStreamFlags|=AUDCLNT_STREAMFLAGS_SRC_DEFAULT_QUALITY;
    DWRITELN("AUDCLNT_STREAMFLAGS_SRC_DEFAULT_QUALITY");
  shared: ;
  }

  DVWRITELN("dwStreamFlags=%0lx",pEngine->param.dwStreamFlags);
  pEngine->param.bDoubleBuffered=pDevice->eMode&&pDevice->eStrategy;

  /////////////////////////////////////////////////////////////////////////////
  switch (pEngine->options.common.eVolume) {
  case eVolumeSoftware:
    pEngine->param.pIEngineVolumeVtbl=&gcIEngineVolumeSoftwareVtbl;
    break;
  case eVolumeWasapi:
    switch (pEngine->options.device.eMode) {
    case eModeShared:
      pEngine->param.pIEngineVolumeVtbl=&gcIEngineVolumeSimpleAudioVtbl;
      break;
    case eModeExclusive:
      pEngine->param.pIEngineVolumeVtbl=&gcIEngineVolumeAudioEndpointVtbl;
      break;
    default:
      pEngine->param.pIEngineVolumeVtbl=&gcIEngineVolumeSoftwareVtbl;
      break;
    }

    break;
  default:
    pEngine->param.pIEngineVolumeVtbl=&gcIEngineVolumeNullVtbl;
    break;
  };

  /////////////////////////////////////////////////////////////////////////////
  code=0;
exit:
  return code;
}


static int EngineGetBufferDuration(Engine *pEngine)
{
  int code=-1;
  REFERENCE_TIME hnsBufferDuration;
  HRESULT hr;

  /////////////////////////////////////////////////////////////////////////////
  if (pEngine->options.device.eMode) {
    hr=pEngine->pAudioClient->lpVtbl->GetDevicePeriod(pEngine->pAudioClient,
      NULL,       // [out] REFERENCE_TIME *phnsDefaultDevicePeriod,
      &hnsBufferDuration
                  // [out] REFERENCE_TIME *phnsMinimumDevicePeriod
    );
  }
  else {
    hr=pEngine->pAudioClient->lpVtbl->GetDevicePeriod(pEngine->pAudioClient,
      &hnsBufferDuration,
                  // [out] REFERENCE_TIME *phnsDefaultDevicePeriod,
      NULL        // [out] REFERENCE_TIME *phnsMinimumDevicePeriod
    );
  }

  if (FAILED(hr)) {
    switch (hr) {
    DCMESSAGE("IAudioClient::GetDevicePeriod",AUDCLNT_E_DEVICE_INVALIDATED);
    DCMESSAGE("IAudioClient::GetDevicePeriod",AUDCLNT_E_SERVICE_NOT_RUNNING);
    DCMESSAGE("IAudioClient::GetDevicePeriod",E_POINTER);
    DDMESSAGE("IAudioClient::GetDevicePeriod",hr);
    }

    goto exit;
  }

  DVWRITELN("*** hnsBufferDuration=%I64d ***",hnsBufferDuration);
  pEngine->param.hnsBufferDuration=hnsBufferDuration;

  /////////////////////////////////////////////////////////////////////////////
  code=0;
exit:
  return code;
}

int EngineCreateAudioClient(Engine *pEngine, const EEngine eState)
{
  int code=-1;
  int nAlignment=0;
  UINT32 uNumFrames;
  HRESULT hr;

  /////////////////////////////////////////////////////////////////////////////
activate:
  hr=pEngine->pDevice->lpVtbl->Activate(pEngine->pDevice,
    &IID_IAudioClient,      // [in]  REFIID      iid,
    CLSCTX_ALL,             // [in]  DWORD       dwClsCtx,
    NULL,                   // [in]  PROPVARIANT *pActivationParams,
    (void **)&pEngine->pAudioClient
                            // [out] void        **ppInterface
  );

  if (FAILED(hr)) {
    switch (hr) {
    DCMESSAGE("IMMDevice::Activate",E_NOINTERFACE);
    DCMESSAGE("IMMDevice::Activate",E_POINTER);
    DCMESSAGE("IMMDevice::Activate",E_INVALIDARG);
    DCMESSAGE("IMMDevice::Activate",E_OUTOFMEMORY);
    DCMESSAGE("IMMDevice::Activate",AUDCLNT_E_DEVICE_INVALIDATED);
    DDMESSAGE("IMMDevice::Activate",hr);
    }

    goto exit;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (0ll==pEngine->param.hnsBufferDuration) {
    if (eEngineAudioClientInititial!=eState) {
      DMESSAGE("internal");
      goto exit;
    }
    else if (EngineGetBufferDuration(pEngine)<0) {
      DMESSAGE("retrieving buffer duration");
      goto exit;
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  if (pEngine->options.device.eStrategy) {
    hr=pEngine->pAudioClient->lpVtbl->Initialize(pEngine->pAudioClient,
      pEngine->param.eShareMode,
                            // [in]       AUDCLNT_SHAREMODE ShareMode,
      pEngine->param.dwStreamFlags,
                            // [in]       DWORD             StreamFlags,
      pEngine->param.hnsBufferDuration,
                            // [in]       REFERENCE_TIME    hnsBufferDuration,
      pEngine->param.hnsBufferDuration,
                            // [in]       REFERENCE_TIME    hnsPeriodicity,
      &pEngine->wfx.target.Format,
                            // [in] const WAVEFORMATEX      *pFormat,
      NULL                  // [in]       LPCGUID           AudioSessionGuid
    );
  }
  else {
    hr=pEngine->pAudioClient->lpVtbl->Initialize(pEngine->pAudioClient,
      pEngine->param.eShareMode,
                            // [in]       AUDCLNT_SHAREMODE ShareMode,
      pEngine->param.dwStreamFlags,
                            // [in]       DWORD             StreamFlags,
      pEngine->param.hnsBufferDuration,
                            // [in]       REFERENCE_TIME    hnsBufferDuration,
      0ull,                 // [in]       REFERENCE_TIME    hnsPeriodicity,
      &pEngine->wfx.target.Format,
                            // [in] const WAVEFORMATEX      *pFormat,
      NULL                  // [in]       LPCGUID           AudioSessionGuid
    );
  }

#define DCLABEL "IAudioClient::Initialize" // [
  if (FAILED(hr)) {
    switch (hr) {
    DCMESSAGE(DCLABEL,CO_E_NOTINITIALIZED);
    DCMESSAGE(DCLABEL,AUDCLNT_E_ALREADY_INITIALIZED);
    DCMESSAGE(DCLABEL,AUDCLNT_E_WRONG_ENDPOINT_TYPE);
    case AUDCLNT_E_BUFFER_SIZE_NOT_ALIGNED:
      if (pEngine->options.device.bAlignmentDance&&!nAlignment) {
        // https://blogs.msdn.microsoft.com/matthew_van_eerde/2009/04/03/sample-wasapi-exclusive-mode-event-driven-playback-app-including-the-hd-audio-alignment-dance/
        hr=pEngine->pAudioClient->lpVtbl->GetBufferSize(pEngine->pAudioClient,
          &uNumFrames                             // UINT32 *pNumBufferFrames
        );

        pEngine->param.hnsBufferDuration=
            10.0*1000                             // (hns / ms) *
            *1000                                 // (ms / s) *
            *uNumFrames                           // frames /
            /pEngine->wfx.target.Format.nSamplesPerSec
                                                  // (frames / s)
            +0.5                                  // rounding
            ;

        DVWRITELN("*** uNumFrames (first pass)=%u ***",uNumFrames);
        DVWRITELN("*** hnsBufferDuration (first pass)=%I64d ***",
            pEngine->param.hnsBufferDuration);

        --pEngine->eState;
        DVWRITELN("--< %s",EngineState(pEngine));
        hr=pEngine->pAudioClient->lpVtbl->Release(pEngine->pAudioClient);
        pEngine->pAudioClient=NULL;
        ++nAlignment;
        goto activate;
      }

      DMESSAGE(DCLABEL " failed with AUDCLNT_E_BUFFER_SIZE_NOT_ALIGNED");
      break;
    DCMESSAGE(DCLABEL,AUDCLNT_E_BUFFER_SIZE_ERROR);
    DCMESSAGE(DCLABEL,AUDCLNT_E_CPUUSAGE_EXCEEDED);
    DCMESSAGE(DCLABEL,AUDCLNT_E_DEVICE_INVALIDATED);
    DCMESSAGE(DCLABEL,AUDCLNT_E_DEVICE_IN_USE);
    case AUDCLNT_E_ENDPOINT_CREATE_FAILED:
      if (1==pEngine->wfx.target.Format.nChannels
          &&!pEngine->options.common.eMono2Stereo) {
        DMESSAGE(DCLABEL " failed with AUDCLNT_E_ENDPOINT_CREATE_FAILED"
            " (consider switching on mono to stereo convertion)");
        goto exit;
      }

      DVMESSAGE(DCLABEL " failed with AUDCLNT_E_ENDPOINT_CREATE_FAILED"
          " (nChannels: %u, nSamplesPerSec: %lu, nBlockAlign: %u,"
          " wBitsPerSample: %u)",
          pEngine->wfx.target.Format.nChannels,
          pEngine->wfx.target.Format.nSamplesPerSec,
          pEngine->wfx.target.Format.nBlockAlign,
          pEngine->wfx.target.Format.wBitsPerSample);
      break;
    DCMESSAGE(DCLABEL,AUDCLNT_E_INVALID_DEVICE_PERIOD);
    case AUDCLNT_E_UNSUPPORTED_FORMAT:
      DVMESSAGE(DCLABEL " failed with AUDCLNT_E_UNSUPPORTED_FORMAT"
          " (nChannels: %u, nSamplesPerSec: %lu, wBitsPerSample: %u)",
          pEngine->wfx.target.Format.nChannels,
          pEngine->wfx.target.Format.nSamplesPerSec,
          pEngine->wfx.target.Format.wBitsPerSample);
      break;
    DCMESSAGE(DCLABEL,AUDCLNT_E_EXCLUSIVE_MODE_NOT_ALLOWED);
    DCMESSAGE(DCLABEL,AUDCLNT_E_BUFDURATION_PERIOD_NOT_EQUAL);
    DCMESSAGE(DCLABEL,AUDCLNT_E_SERVICE_NOT_RUNNING);
    DCMESSAGE(DCLABEL,E_POINTER);
    DCMESSAGE(DCLABEL,E_INVALIDARG);
    DCMESSAGE(DCLABEL,E_OUTOFMEMORY);
    DDMESSAGE(DCLABEL,hr);
    }

    goto exit;
  }
#undef DCLABEL // ]

  /////////////////////////////////////////////////////////////////////////////
  if (0u==pEngine->param.uNumFramesDevice) {
    if (eEngineAudioClientInititial==eState) {
      hr=pEngine->pAudioClient->lpVtbl->GetBufferSize(pEngine->pAudioClient,
        &uNumFrames       // UINT32 *pNumBufferFrames
      );

#define DCLABEL "IAudioClient::GetBufferSize" // [
      if (FAILED(hr)) {
        switch (hr) {
        DCMESSAGE(DCLABEL,AUDCLNT_E_NOT_INITIALIZED);
        DCMESSAGE(DCLABEL,AUDCLNT_E_DEVICE_INVALIDATED);
        DCMESSAGE(DCLABEL,AUDCLNT_E_SERVICE_NOT_RUNNING);
        DCMESSAGE(DCLABEL,E_POINTER);
        DDMESSAGE(DCLABEL,hr);
        }

        goto exit;
      }
#undef DCLABEL // ]

      pEngine->param.uNumFramesDevice=uNumFrames;

      if (!pEngine->options.device.eStrategy) {
        pEngine->param.DueTime.QuadPart
            =-0.5*YASAPI_FRAMES2HNS(uNumFrames,&pEngine->wfx.target)-0.5;

        if (pEngine->options.device.bTimerPeriodic) {
          pEngine->param.lPeriod
              =0.5*YASAPI_FRAMES2MS(uNumFrames,&pEngine->wfx.target)+0.5;
        }
        else
          pEngine->param.lPeriod=0l;
      }
      else {
        pEngine->param.DueTime.QuadPart=0ll;
        pEngine->param.lPeriod=0l;
      }
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  if (eStrategyPull==pEngine->options.device.eStrategy) {
    hr=pEngine->pAudioClient->lpVtbl->SetEventHandle(pEngine->pAudioClient,
      pEngine->timer.hEvent   // [in] HANDLE eventHandle
    );

    if (FAILED(hr)) {
      switch (hr) {
      DCMESSAGE("IAudioClient::SetEventHandle",E_INVALIDARG);
      DCMESSAGE("IAudioClient::SetEventHandle",
          AUDCLNT_E_EVENTHANDLE_NOT_EXPECTED);
      DCMESSAGE("IAudioClient::SetEventHandle",AUDCLNT_E_NOT_INITIALIZED);
      DCMESSAGE("IAudioClient::SetEventHandle",AUDCLNT_E_DEVICE_INVALIDATED);
      DCMESSAGE("IAudioClient::SetEventHandle",AUDCLNT_E_SERVICE_NOT_RUNNING);
      DDMESSAGE("IAudioClient::SetEventHandle",hr);
      }
  
      EngineKill(pEngine,eState);
      goto exit;
    }

    DWRITELN("event handle registered");
  }

  /////////////////////////////////////////////////////////////////////////////
  code=0;
exit:
  return code;
}

///////////////////////////////////////////////////////////////////////////////
void EngineKill(Engine *pEngine, const EEngine eState)
{
  Gtk *pGtk=&pEngine->pPlayer->gtk;

  switch (pEngine->eState) {
  case eEngineEot:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineEot==pEngine->eState,exit);
    --pEngine->eState;
    // Intentional fall-through.
  case eEngineDraining:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineDraining==pEngine->eState,exit);
    --pEngine->eState;
    // Intentional fall-through.
  case eEngineTimerStarted:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineTimerStarted==pEngine->eState,exit);
    --pEngine->eState;
    CancelWaitableTimer(pEngine->timer.hTimer);
    // Intentional fall-through.
  case eEngineAudioClientStarted:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineAudioClientStarted==pEngine->eState,exit);
    --pEngine->eState;
    pEngine->pAudioClient->lpVtbl->Stop(pEngine->pAudioClient);

    if (pEngine->param.bDoubleBuffered)
      pEngine->pAudioClient->lpVtbl->Reset(pEngine->pAudioClient);
    // Intentional fall-through.
  case eEnginePauseConnected:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEnginePauseConnected==pEngine->eState,exit);
    --pEngine->eState;
    // Intentional fall-through.
  case eEngineAudioRenderClient:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineAudioRenderClient==pEngine->eState,exit);
    --pEngine->eState;
    pEngine->pAudioRenderClient->lpVtbl->Release(pEngine->pAudioRenderClient);
    pEngine->pAudioRenderClient=NULL;
    // Intentional fall-through.
  case eEngineAudioClock:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineAudioClock==pEngine->eState,exit);

    pEngine->param.time.lpVtbl->ClockDestroy(pEngine);

    --pEngine->eState;
    // Intentional fall-through.
  case eEngineVolume:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineVolume==pEngine->eState,exit);
    --pEngine->eState;
    pEngine->param.pIEngineVolumeVtbl->Destroy(pEngine);
    // Intentional fall-through.
  case eEngineAudioClient:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineAudioClient==pEngine->eState,exit);
    --pEngine->eState;
    pEngine->pAudioClient->lpVtbl->Release(pEngine->pAudioClient);
    pEngine->pAudioClient=NULL;
    // Intentional fall-through.
  case eEnginePauseDisconnected:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEnginePauseDisconnected==pEngine->eState,exit);
    --pEngine->eState;
    // Intentional fall-through.
  case eEngineRing:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineRing==pEngine->eState,exit);
    --pEngine->eState;
    pEngine->monitor.xWritten=0.0;
    pEngine->monitor.xPadding=0.0;
    RingDestroy(&pEngine->ring);

    ZeroMemory(&pEngine->param,sizeof pEngine->param);
    ZeroMemory(&pEngine->stats,sizeof pEngine->stats);
    pEngine->stats.uNumTracks=UINT32_MAX;
    pEngine->stats.uNumDropouts=UINT32_MAX;
    pEngine->param.uFramesPlaybackBegin=0u;
    pEngine->monitor.xStart=0.0;
    GtkUpdateDialog(pGtk,&pGtk->config);
    // Intentional fall-through.
  case eEngineAudioClientInititial:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineAudioClientInititial==pEngine->eState,exit);
    --pEngine->eState;

    if (pEngine->pAudioClient) {
      pEngine->pAudioClient->lpVtbl->Release(pEngine->pAudioClient);
      pEngine->pAudioClient=NULL;
    }
    // Intentional fall-through.
  case eEngineThreadRunning:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineThreadRunning==pEngine->eState,exit);
    --pEngine->eState;
    EngineSend(pEngine,eCommandExit,eEngineNull,1);
    WaitForSingleObject(pEngine->thread.hThread,INFINITE);
    // Intentional fall-through.
  case eEngineThread:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineThread==pEngine->eState,exit);
    --pEngine->eState;
    CloseHandle(pEngine->thread.hThread);
    pEngine->thread.hThread=NULL;
    // Intentional fall-through.
  case eEngineEventConsumer:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineEventConsumer==pEngine->eState,exit);
    --pEngine->eState;
    CloseHandle(pEngine->thread.hConsumer);
    pEngine->thread.hConsumer=NULL;
    // Intentional fall-through.
  case eEngineEventProducer:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineEventProducer==pEngine->eState,exit);
    --pEngine->eState;
    CloseHandle(pEngine->thread.hProducer);
    pEngine->thread.hProducer=NULL;
    // Intentional fall-through.
  case eEngineTimer:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineTimer==pEngine->eState,exit);
    --pEngine->eState;
    CloseHandle(pEngine->timer.hTimer);
    pEngine->timer.hTimer=NULL;
    // Intentional fall-through.
  case eEngineTimerEvent:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineTimerEvent==pEngine->eState,exit);
    --pEngine->eState;
    CloseHandle(pEngine->timer.hEvent);
    pEngine->timer.hEvent=NULL;
    // Intentional fall-through.
  case eEngineCriticalSection:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineCriticalSection==pEngine->eState,exit);
    --pEngine->eState;
    DeleteCriticalSection(&pEngine->cs);
    // Intentional fall-through.
  default:
    break;
  }

  DWRITELN(EngineState(pEngine));
DEXIT();
  return;
}

void EngineDestroy(Engine *pEngine)
{
  EngineKill(pEngine,eEngineNull);
}

void EngineClose(Engine *pEngine)
{

DVWRITELN("### %s ####################################",__func__);
  if (pEngine->eState<eEngineBase) {
    DMESSAGE("not initialized");
    goto enoinit;
  }

  EnterCriticalSection(&pEngine->cs); // {
  DVWRITELN("=== target: %u/%u/%u ===",
      pEngine->wfx.target.Format.nSamplesPerSec,
      pEngine->wfx.target.Format.nChannels,
      pEngine->wfx.target.Format.wBitsPerSample);
    EngineSend(pEngine,eCommandKill,eEngineBase,0);
  DVWRITELN("=== target: %u/%u/%u ===",
      pEngine->wfx.target.Format.nSamplesPerSec,
      pEngine->wfx.target.Format.nChannels,
      pEngine->wfx.target.Format.wBitsPerSample);
  LeaveCriticalSection(&pEngine->cs); // }
enoinit: ;
}

///////////////////////////////////////////////////////////////////////////////
int EnginePause(Engine *pEngine, int nPause)
{
  int nLastPause=pEngine->param.nLastPause;

  if (nLastPause==nPause)
    goto epause;

  if (pEngine->eState<eEngineBase)
    goto ebase;

  EnterCriticalSection(&pEngine->cs); // {

  if (nPause) {
    switch (pEngine->eState) {
    case eEngineStarted:
      if (pEngine->options.common.bDisconnect)
        EngineSend(pEngine,eCommandKill,eEnginePauseDisconnected,0);
      else
        EngineSend(pEngine,eCommandKill,eEnginePauseConnected,0);

      pEngine->param.nLastPause=nPause;
      break;
    default:
      break;
    }
  }
  else {
    switch (pEngine->eState) {
    case eEnginePauseDisconnected:
    case eEnginePauseConnected:
      EngineSend(pEngine,eCommandStart,eEngineNull,0);
      pEngine->param.nLastPause=nPause;
      break;
    default:
      break;
    }
  }

  LeaveCriticalSection(&pEngine->cs); // }
ebase:
epause:
  return nLastPause;
}

int EngineIsPlaying(Engine *pEngine)
{
  int bIsPlaying=0;

  if (pEngine->eState<eEngineBase)
    goto ebase;

  EnterCriticalSection(&pEngine->cs); // {

  switch (pEngine->eState) {
  case eEnginePauseDisconnected:
  case eEnginePauseConnected:
    break;
  case eEngineStarted:
    ++pEngine->eState;
    DASSERT(eEngineDraining==pEngine->eState,exit);
    bIsPlaying=1;
    break;
  case eEngineDraining:
    bIsPlaying=1;
    break;
  default:
    break;
  }
DEXIT();
  LeaveCriticalSection(&pEngine->cs); // }
ebase:
  return bIsPlaying;
}

int EngineAvailable(Engine *pEngine)
{
  Ring *pRing=&pEngine->ring;
  int nAvailable=0;

  if (pEngine->eState<eEngineBase)
    goto ebase;

  EnterCriticalSection(&pEngine->cs); // {


  switch (pEngine->eState) {
  case eEnginePauseDisconnected:
  case eEnginePauseConnected:
    break;
  default:
    if (pEngine->eState<eEngineRing)
      nAvailable=INT_MAX;
    else
      nAvailable=RingSourceSizeAvailable(pRing);

    break;
  }

  LeaveCriticalSection(&pEngine->cs); // }
ebase:
//DVWRITELN("nAvailable: %d",nAvailable);
  return nAvailable;
}


// Return time history corrected.
int EngineOutputTime(Engine *pEngine)
{
  int ms=0;

  if (pEngine->options.common.bLatency&&pEngine->eState<eEngineBase)
    goto ebase;

  EnterCriticalSection(&pEngine->cs); // {

  switch (pEngine->eState) {
  case eEnginePauseDisconnected:
  case eEnginePauseConnected:
  case eEngineEot:
    goto exit;
  default:
    if (pEngine->eState<eEngineAudioClock)
      goto exit;

    break;
  }

  ms=pEngine->param.time.lpVtbl->Get(pEngine);
exit:
  LeaveCriticalSection(&pEngine->cs); // }
ebase:
  return ms;
}

int EngineWrittenTime(Engine *pEngine)
{
  WAVEFORMATEXTENSIBLE *pTarget=&pEngine->wfx.target;
  double msWrittenTime=0.0;
  Ring *pRing=&pEngine->ring;
  WORD uFramesWritten;
  UINT32 uNumFramesPadding;

  if (pEngine->eState<eEngineBase)
    goto ebase;

  EnterCriticalSection(&pEngine->cs); // {

  switch (pEngine->eState) {
  case eEnginePauseDisconnected:
  case eEnginePauseConnected:
  case eEngineEot:
    goto exit;
  default:
    if (pEngine->eState<eEngineAudioClock)
      goto exit;

    break;
  }

  uFramesWritten=pRing->frames._uWritten;
  msWrittenTime=YASAPI_FRAMES2MS(uFramesWritten,pTarget);

  msWrittenTime-=pEngine->param.time.lpVtbl->Get(pEngine);

  if (!pEngine->options.common.bLatency&&!pEngine->param.bDoubleBuffered) {
    if (EngineNumFramesPadding(pEngine,&uNumFramesPadding)<0)
      goto exit;

    msWrittenTime+=YASAPI_FRAMES2MS(uNumFramesPadding,&pEngine->wfx.target);
  }
exit:
  LeaveCriticalSection(&pEngine->cs); // }
ebase:
  return msWrittenTime+0.5;
}

static void EngineCeilFrames(UINT32 *puFrames, UINT32 uChunkFrames)
{
  UINT32 uMod=*puFrames%uChunkFrames;

  if (uMod)
    *puFrames+=uChunkFrames-uMod;
}

static int EngineCreateRing(Engine *pEngine, DWORD dwChunkSizeSource,
    ERingAlloc eRingRealloc)
{
  int code=-1;
  Gtk *pGtk=&pEngine->pPlayer->gtk;
  EEngine eState=pEngine->eState;
  Ring *pRing=&pEngine->ring;
  ////////////////
  WAVEFORMATEX *pSource=&pEngine->wfx.source.Format;
  UINT32 uFramesChunk=dwChunkSizeSource/pSource->nBlockAlign;
  double msLatency=pEngine->options.device.xMaxLatencyAssumed;
  UINT32 uFramesRing=YASAPI_MS2FRAMES(msLatency,&pEngine->wfx.target)+0.5;
  ////////////////
  double xPlaybackBegin=pEngine->options.device.xPlaybackBegin;
  UINT32 uFramesDevice=pEngine->param.uNumFramesDevice;
  UINT32 uFramesPlaybackBegin=xPlaybackBegin*uFramesDevice+0.5;

  /////////////////////////////////////////////////////////////////////////////
  pEngine->param.uFramesPlaybackBegin=uFramesPlaybackBegin;

DVWRITELN("uFramesRing: %u ====================================",uFramesRing);
  if (uFramesRing<uFramesPlaybackBegin)
    uFramesRing=uFramesPlaybackBegin;

DVWRITELN("uFramesRing: %u ====================================",uFramesRing);
  /////////////////////////////////////////////////////////////////////////////
  EngineCeilFrames(&uFramesRing,uFramesChunk);
DVWRITELN("uFramesRing: %u (%u %u) =============================",uFramesRing,
  uFramesRing/uFramesChunk,uFramesRing%uFramesChunk);

  pEngine->param.uNumFramesRing=uFramesRing;
  GtkUpdateWidget(pGtk,&pGtk->config.config.monitor.engine.num_frames_ring);

  if (eRingRealloc) {
    if (RingReallocFrames(pRing,pEngine,uFramesRing)<0) {
      DMESSAGE("reallocating ring");
      EngineKill(pEngine,eState);
      goto exit;
    }
  }
  else {
    if (RingCreateFrames(pRing,pEngine,uFramesRing)<0) {
      DMESSAGE("creating ring");
      EngineKill(pEngine,eState);
      goto exit;
    }

    ++pEngine->eState;
    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineRing==pEngine->eState,exit);
  }

  pEngine->param.dwChunkSizeSource=dwChunkSizeSource;

  pEngine->monitor.xStart=(double)uFramesPlaybackBegin/uFramesRing;
  GtkUpdateWidget(pGtk,&pGtk->config.config.monitor.start);

  /////////////////////////////////////////////////////////////////////////////
  code=0;
exit:
  return code;
}

static int EngineChunkSizeKnown(Engine *pEngine, DWORD dwChunkSizeSource)
{
   return pEngine->param.dwChunkSizeSource==dwChunkSizeSource;
}

int EngineWrite(Engine *pEngine, BYTE *buf, DWORD dwChunkSizeSource)
{
  int code=-1;
  Player *pPlayer=pEngine->pPlayer;
  Gtk *pGtk=&pPlayer->gtk;
  Ring *pRing=&pEngine->ring;
  WORD uFramesPlaybackBegin;
  WORD uFramesWritten;
  WORD uFramesMax;
  DWORD dwSourceSize;
  UINT32 uMsLatency;

  if (pEngine->eState<eEngineBase) {
    DMESSAGE("engine not initialized");
    goto ebase;
  }

  EnterCriticalSection(&pEngine->cs); // {

  switch (pEngine->eState) {
  case eEngineDraining:
    // We're still in state "eEngineDraining" ...
    EngineKill(pEngine,eEngineTimerStarted);
    ++pEngine->stats.uNumTracks;
    ++pEngine->pPlayer->stats.uNumTracks;

    if (EngineCreateRing(pEngine,dwChunkSizeSource,eRingRealloc)<0) {
      DMESSAGE("creating ring");
      goto exit;
    }

    GtkUpdateDialog(&pPlayer->gtk,&pPlayer->gtk.config);
    break;
  default:
    if (pEngine->eState<eEngineRing) {
      if (EngineCreateRing(pEngine,dwChunkSizeSource,eRingCreate)<0) {
        DMESSAGE("creating ring");
        goto exit;
      }
    }
    else if (!EngineChunkSizeKnown(pEngine,dwChunkSizeSource)) {
      // This is special to "in_mp3" because it has not a constant block size.
      dwSourceSize=RingSourceSizeAvailable(pRing)+dwChunkSizeSource;

      if (RingSourceSize(pRing)<dwSourceSize
          &&EngineCreateRing(pEngine,dwChunkSizeSource,eRingRealloc)<0) {
        DMESSAGE("creating ring");
        goto exit;
      }
    }

    break;
  }

  /////////////////////////////////////////////////////////////////////////////
  switch (pEngine->eState) {
  case eEnginePauseDisconnected:
  case eEnginePauseConnected:
    DWRITELN("Warning: pause");
    goto exit;
  default:
    break;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (pEngine->eState<eEngineRing) {
    DASSERT_ENGINE_STATE(pEngine,eEngineAudioClientInititial,exit);

    if (EngineCreateRing(pEngine,dwChunkSizeSource,eRingCreate)<0) {
      DMESSAGE("creating ring");
      goto exit;
    }
  }

  uFramesPlaybackBegin=pEngine->param.uFramesPlaybackBegin;

  if (RingWrite(pRing,buf,dwChunkSizeSource)<0) {
    DMESSAGE("writing to ring");
    goto exit;
  }

  /////////////////////////////////////////////////////////////////////////////
  uFramesWritten=pRing->frames._uWritten;

  if (pEngine->stats.uNumFramesMax<uFramesWritten) {
    pEngine->stats.uNumFramesMax=uFramesWritten;
    GtkUpdateWidget(pGtk,
        &pGtk->config.config.monitor.engine.num_frames_latency);
  }

  /////////////////////////////////////////////////////////////////////////////
  uMsLatency=YASAPI_FRAMES2MS(uFramesWritten,&pEngine->wfx.source)+0.5;

  if (pPlayer->stats.uMsLatencyMax<uMsLatency) {
    pPlayer->stats.uMsLatencyMax=uMsLatency;
    GtkUpdateWidget(pGtk,
        &pGtk->config.config.monitor.player.ms_latency);
  }

  /////////////////////////////////////////////////////////////////////////////
  if (eEngineStarted<=pEngine->eState)
    goto success;

  if (uFramesWritten<uFramesPlaybackBegin) {
    ///////////////////////////////////////////////////////////////////////////
    uMsLatency=YASAPI_FRAMES2MS(uFramesWritten,&pEngine->wfx.source)+0.5;


    DVWRITELN("*** underflow: %u (%u %u)",uFramesWritten,uFramesPlaybackBegin,
        dwChunkSizeSource);
    //uFramesMax=pRing->lpVtbl->FramesMaximum(pRing);
    uFramesMax=RingFramesMaximum(pRing);
    pEngine->monitor.xWritten=(double)uFramesWritten/uFramesMax;
    GtkUpdateWidget(pGtk,&pGtk->config.config.monitor.written);
    goto success;
  }

  DVWRITELN("*** starting: %u (%u)",uFramesWritten,uFramesPlaybackBegin);
  EngineSend(pEngine,eCommandStart,eEngineNull,0);
  DWRITELN(EngineState(pEngine));
success:
  code=0;
exit:
  LeaveCriticalSection(&pEngine->cs); // }
ebase:
  return code;
}



void EngineVolume(Engine *pEngine, Player *pPlayer)
{
  if (pEngine->eState<eEngineBase&&EngineCreateBase(pEngine,pPlayer)<0) {
    DWRITELN("engine not initialized");
    goto ebase;
  }

  EnterCriticalSection(&pEngine->cs); // {

  if (pEngine->eState<eEngineStarted) {
    DWRITELN(EngineState(pEngine));
    goto exit;
  }

  pEngine->param.pIEngineVolumeVtbl->Set(pEngine);
exit:
  LeaveCriticalSection(&pEngine->cs); // }
ebase:
  return;
}



///////////////////////////////////////////////////////////////////////////////
void EngineSend(Engine *pEngine, const ECommand eCommand, EEngine eArgument,
  int bLock)
{
  Command command={ eCommand, eArgument };

  if (bLock)
    EnterCriticalSection(&pEngine->cs); // {

  pEngine->thread.pCommand=&command;
  SetEvent(pEngine->thread.hProducer);

  while (pEngine->thread.pCommand) {
    LeaveCriticalSection(&pEngine->cs); // {
    WaitForSingleObject(pEngine->thread.hConsumer,INFINITE);
    EnterCriticalSection(&pEngine->cs); // }
  }

  if (bLock)
    LeaveCriticalSection(&pEngine->cs); // }
}
